@@ -1,28 +1,23 @@
 package Apache;
 use strict;
-use vars qw($VERSION @ISA @EXPORT_OK);
-use Apache::Constants qw(OK);
 use DynaLoader ();
+use Apache::Constants ();
 
-@ISA = qw(Exporter DynaLoader);
-@EXPORT_OK = qw(exit warn);
-$VERSION = "1.13";
-$Apache::CRLF = "\015\012";
+@Apache::ISA = qw(Exporter DynaLoader);
+@Apache::EXPORT_OK = qw(exit warn);
+$Apache::VERSION = "1.17";
 
-bootstrap Apache $VERSION;
-
-*Apache::warn = \&Apache::log_error;
-
-sub exit {
-    my($r, $s);
-    if(scalar @_ > 1) {
-	($r,$s) = @_;
-    }
-    else {
-	($r,$s) = (Apache->request, @_);
-    }
-    $r->rflush;
-    $r->__exit($s);
+if (caller eq "CGI::Apache") {
+    #we must die here outside of httpd so CGI::Switch works
+    bootstrap Apache $Apache::VERSION;
+}
+else {
+    eval { bootstrap Apache $Apache::VERSION; } 
+}
+if($@) {
+    my $gw = $ENV{GATEWAY_INTERFACE} || '';
+    die "$@\n" if substr($gw,0,8) eq 'CGI-Perl';
+    warn "warning: can't `bootstrap Apache $Apache::VERSION' outside of httpd\n";
 }
 
 sub module {
@@ -66,14 +61,17 @@ sub read {
     $nrd = $total = 0;
     $buf = "";
     $_[1] ||= "";
+    #$#{$_[1]} = $bufsiz unless defined $_[1];
+
     $r->hard_timeout("Apache->read");
 
     while($bufsiz) {
 	$nrd = $r->read_client_block($buf, $bufsiz) || 0;
 	if(defined $nrd and $nrd > 0) {
 	    $bufsiz -= $nrd;
-	    $total += $nrd;
 	    $_[1] .= $buf;
+ 	    #substr($_[1], $total, $nrd) = $buf;
+	    $total += $nrd;
 	    next if $bufsiz;
 	    last;
 	}
@@ -97,6 +95,12 @@ sub READLINE {
 
 *PRINT = \&print;
 
+sub PRINTF {
+    my $r = shift;
+    my $fmt = shift;
+    $r->print(sprintf($fmt, @_));
+}
+
 sub send_cgi_header {
     my($r, $headers) = @_;
     my $dlm = "\015?\012"; #a bit borrowed from LWP::UserAgent
@@ -107,11 +111,8 @@ sub send_cgi_header {
 	if (/^(\S+?):\s*(.*)$/) {
 	    ($key, $val) = ($1, $2);
 	    last unless $key;
-	    if($key eq "Status") {
-		$r->status_line($val);
-		next;
-	    }
-	    elsif($key eq "Location") {
+	    $r->cgi_header_out($key, $val);
+	    if($key eq "Location") {
 		if($val =~ m,^/,) {
 		    #/* This redirect needs to be a GET no 
                     #   matter what the original
@@ -120,22 +121,8 @@ sub send_cgi_header {
 		    $r->method("GET");
 		    $r->method_number(0); #M_GET 
 		    $r->internal_redirect_handler($val);
-		    return OK;
+		    return 0;
 		}
-		else {
-		    $r->header_out(Location => $val);
-		    $r->err_header_out(Location => $val);
-		    $r->status(302);
-		    next;
-		}
-	    }
-	    elsif($key eq "Content-type") {
-		$r->content_type($val);
-		next;
-	    }
-	    else {
-		$r->header_out($key,$val);
-		next;
 	    }
 	}
 	else {
@@ -181,10 +168,6 @@ sub TIEHANDLE {
     $r ||= Apache->request;
 }
 
-#some backwards-compatible stuff
-*Apache::TieHandle::TIEHANDLE = \&TIEHANDLE;
-*CGI::Apache::exit = \&Apache::exit;
-
 1;
 
 __END__
@@ -212,7 +195,7 @@ request object as parameter and may choose update or use it in various
 ways.  Most of the methods described below obtain information from or
 updates the request object.
 The perl version of the request object will be blessed into the B<Apache> 
-package, it is really a C<request_rec *> in disguise.
+package, it is really a C<request_rec*> in disguise.
 
 =over 4
 
@@ -222,45 +205,83 @@ The Apache->request method will return a reference to the request object.
 
 B<Perl*Handler>s can obtain a reference to the request object when it
 is passed to them via C<@_>.  However, scripts that run under 
-L<Apache::Registry>, for example, need a way to access the request object.
-L<Apache::Registry> will make a request object availible to these scripts
+B<Apache::Registry>, for example, need a way to access the request object.
+B<Apache::Registry> will make a request object available to these scripts
 by passing an object reference to C<Apache-E<gt>request($r)>.
-If handlers use modules such as C<Apache::CGI> that need to access
-L<Apache-E<gt>request>, they too should do this (e.g. Apache::Status).
+If handlers use modules such as B<CGI::Apache> that need to access
+C<Apache-E<gt>request>, they too should do this (e.g. B<Apache::Status>).
 
 =item $r->as_string
 
-Returns a string representation of the request object.
+Returns a string representation of the request object.  Mainly useful
+for debugging.
 
 =item $r->main
 
-If the current request is a sub-request, this method returns a blessed 
-reference to the main request structure.
+If the current request is a sub-request, this method returns a blessed
+reference to the main request structure.  If the current request is
+the main request, then this method returns C<undef>.
 
 =item $r->prev
 
 This method returns a blessed reference to the previous (internal) request
-structure.
+structure or C<undef> if there is no previous request.
 
 =item $r->next
 
 This method returns a blessed reference to the next (internal) request
-structure.
+structure or C<undef> if there is no next request.
 
 =item $r->is_main
 
 Returns true if the current request object is for the main request.
+(Should give the same result as C<!$r-E<gt>main>, but will be more efficient.)
 
 =item $r->is_initial_req
 
-Returns true if the current request is the first internal request, 
-returns false if the request is a sub-request or interal redirect.  
+Returns true if the current request is the first internal request,
+returns false if the request is a sub-request or internal redirect.
+
+=back
+
+=head1 SUB REQUESTS
+
+Apache provides a sub-request mechanism to lookup a uri or filename,
+performing all access checks, etc., without actually running the
+response phase of the given request.  Notice, we have dropped the
+C<sub_req_> prefix here.  The C<request_rec*> returned by the lookup
+methods is blessed into the B<Apache::SubRequest> class.  This way,
+C<destroy_sub_request()> is called automatically during
+C<Apache::SubRequest-E<gt>DESTROY> when the object goes out of scope.  The
+B<Apache::SubRequest> class inherits all the methods from the
+B<Apache> class.
+
+=over 4
+
+=item $r->lookup_uri($uri)
+
+   my $subr = $r->lookup_uri($uri);
+   my $filename = $subr->filename;
+
+   unless(-e $filename) {
+       warn "can't stat $filename!\n";
+   } 
+
+=item $r->lookup_file($filename)
+
+   my $subr = $r->lookup_file($filename);
+
+=item $subr->run
+
+   if($subr->run != OK) {
+       $subr->log_error("something went wrong!");
+   }
 
 =back
 
 =head1 CLIENT REQUEST PARAMETERS
 
-First we will take a look at various methods that can be used to
+In this section we will take a look at various methods that can be used to
 retrieve the request parameters sent from the client.
 In the following examples, B<$r> is a request object blessed into the 
 B<Apache> class, obtained by the first parameter passed to a handler subroutine
@@ -277,9 +298,14 @@ Passing an argument will set the method, mainly used for internal redirects.
 =item $r->method_number( [$num] )
 
 The $r->method_number method will return the request method number.
-Each number corresponds to a string representation such as 
-"GET", "HEAD" or "POST".
-Passing an argument will set the method_number, mainly used for internal redirects and testing authorization restriction masks.
+The method numbers are defined by the M_GET, M_POST,... constants
+available from the B<Apache::Constants> module.  Passing an argument
+will set the method_number, mainly used for internal redirects and
+testing authorization restriction masks.
+
+=item $r->bytes_sent
+
+The number of bytes sent to the client, handy for logging, etc.
 
 =item $r->the_request
 
@@ -326,8 +352,8 @@ string>.  When called in a scalar context, the entire string is
 returned.  When called in a list context, a list of parsed I<key> =>
 I<value> pairs are returned, i.e. it can be used like this:
 
-  $query = $r->args;
-  %in    = $r->args;
+   $query = $r->args;
+   %in    = $r->args;
 
 =item $r->headers_in
 
@@ -340,9 +366,8 @@ header value directly.
 
 Return the value of a client header.  Can be used like this:
 
-  $ct = $r->header_in("Content-type");
-
-  $r->header_in($key, $val); #set the value of header '$key'
+   $ct = $r->header_in("Content-type");
+   $r->header_in($key, $val); #set the value of header '$key'
 
 =item $r->content
 
@@ -358,7 +383,7 @@ as the entire body is read from the client.
 
 Read from the entity body sent by the client.  Example of use:
 
-  $r->read_client_block($buf, $r->header_in('Content-length'));
+   $r->read_client_block($buf, $r->header_in('Content-length'));
 
 =item $r->read($buf, $bytes_to_read)
 
@@ -419,12 +444,42 @@ The dotted decimal representation of the remote client's IP address.
 This is set by then server when the connection record is created so
 is always defined.
 
+=item $c->local_addr
+
+A packed SOCKADDR_IN in the same format as returned by
+L<Socket/pack_sockaddr_in>, containing the port and address on the
+local host that the remote client is connected to.  This is set by
+the server when the connection record is created so it is always
+defined.
+
+=item $c->remote_addr
+
+A packed SOCKADDR_IN in the same format as returned by
+L<Socket/pack_sockaddr_in>, containing the port and address on the
+remote host that the server is connected to.  This is set by the
+server when the connection record is created so it is always defined.
+
+Among other things, this can be used, together with C<$c-E<gt>local_addr>, to
+perform RFC1413 ident lookups on the remote client even when the
+configuration directive B<IdentityCheck> is turned off.
+
+Can be used like:
+
+   use Net::Ident qw (lookupFromInAddr);
+   ...
+   my $remoteuser = lookupFromInAddr ($c->local_addr,
+                                      $c->remote_addr, 2);
+
+Note that the lookupFromInAddr interface does not currently exist in
+the B<Net::Ident> module, but the author is planning on adding it
+soon.
+
 =item $c->remote_logname
 
 If the configuration directive B<IdentityCheck> is set to on:  then the
 first time C<$r-E<gt>get_remote_logname> is called the server does an RFC
 1413 (ident) lookup to get the remote user's system name. Generally for
-UNI* systems this is their login. The result is cached in C<$c->remote_logname>
+UNI* systems this is their login. The result is cached in C<$c-E<gt>remote_logname>
 then returned.  Subsequent calls to C<$r-E<gt>get_remote_host> return the
 cached value.
 
@@ -462,11 +517,11 @@ configuration and access control files.
 Returns the value of a per-directory variable specified by the 
 C<PerlSetVar> directive.
 
- #<Location /foo/bar>
- #SetPerlVar  Key  Value
- #</Location>
+   # <Location /foo/bar>
+   # SetPerlVar  Key  Value
+   # </Location>
 
- my $val = $r->dir_config('Key');
+   my $val = $r->dir_config('Key');
 
 =item $r->requires
 
@@ -503,19 +558,19 @@ Returns a reference to the current value of the per server
 configuration directive B<DocumentRoot>. To quote the Apache server
 documentation, "Unless matched by a directive like Alias, the server
 appends the path from the requested URL to the document root to make
-the path to the document." This same value is passed to CGI
+the path to the document."  This same value is passed to CGI
 scripts in the C<DOCUMENT_ROOT> environment variable.
 
 =item $r->allow_options
 
-The $r->allow_options method can be used for
-checking if it is ok to run a perl script.  The B<Apache::Options>
-module provide the constants to check against.
+The C<$r-E<gt>allow_options> method can be used for
+checking if it is OK to run a perl script.  The B<Apache::Options>
+module provides the constants to check against.
 
- if(!($r->allow_options & OPT_EXECCGI)) {
-     $r->log_reason("Options ExecCGI is off in this directory", 
-		    $filename);
- }
+   if(!($r->allow_options & OPT_EXECCGI)) {
+       $r->log_reason("Options ExecCGI is off in this directory", 
+		      $filename);
+   }
 
 =item $s = $r->server
 
@@ -535,13 +590,13 @@ Returns the hostname used by this server.
 
 Returns the port that this servers listens too.
 
-=item $s->is_virtal
+=item $s->is_virtual
 
 Returns true if this is a virtual server.
 
 =item $s->names
 
-Returns the wildcarded names for HostAlias servers. 
+Returns the wild-carded names for HostAlias servers. 
 
 =back
 
@@ -572,7 +627,7 @@ If the current request is protected by Basic authentication,
 this method will return 0, otherwise -1.  
 The second return value will be the decoded password sent by the client.
 
-    ($ret, $sent_pw) = $r->get_basic_auth_pw;
+   ($ret, $sent_pw) = $r->get_basic_auth_pw;
 
 =item $r->note_basic_auth_failure
 
@@ -585,7 +640,7 @@ for the realm defined by the configuration directive C<AuthName>.
 Set the handler for a request.
 Normally set by the configuration directive C<AddHandler>.
   
- $r->handler( "perl-script" );
+   $r->handler( "perl-script" );
 
 =item $r->notes( $key, [$value] )
 
@@ -595,7 +650,6 @@ modules to pass messages amongst themselves. Generally if you are
 writing handlers in mod_perl you can use Perl variables for this.
 
    $r->notes("MY_HANDLER", OK);
-
    $val = $r->notes("MY_HANDLER");
 
 =item $r->content_type( [$newval] )
@@ -627,7 +681,7 @@ B<Apache::Constants> module provide mnemonic names for the status codes.
 =item $r->status_line( $string )
 
 Get or set the response status line.  The status line is a string like
-"HTTP/1.0 200 OK" and it will take precedence over the value specified
+"200 Document follows" and it will take precedence over the value specified
 using the $r->status() described above.
 
 
@@ -645,7 +699,6 @@ should not define any "Content-XXX" headers by calling this method,
 because these headers use their own specific methods.  Example of use:
 
    $r->header_out("WWW-Authenticate" => "Basic");
-
    $val = $r->header_out($key);
 
 =item $r->err_headers_out
@@ -665,7 +718,6 @@ Change the value of an error response header, or create a new one.
 These headers are used if the status indicates an error.
 
    $r->err_header_out("Warning" => "Bad luck");
-
    $val = $r->err_header_out($key);
 
 =item $r->no_cache( $boolean )
@@ -673,7 +725,7 @@ These headers are used if the status indicates an error.
 This is a flag that indicates that the data being returned is volatile
 and the client should be told not to cache it.
 
-=item $r->print()
+=item $r->print( @list )
 
 This method sends data to the client with C<$r-E<gt>write_client>, but first
 sets a timeout before sending with C<$r-E<gt>hard_timeout>.
@@ -692,7 +744,7 @@ like this:
 Redirect to a location in the server namespace without 
 telling the client. For instance:
 
-  $r->internal_redirect_handler("/home/sweet/home.html");
+   $r->internal_redirect_handler("/home/sweet/home.html");
 
 =back
 
@@ -741,6 +793,16 @@ kill_timeout() will disarm either variety of timeout.
 
 reset_timeout() resets the timeout in progress.
 
+=item $r->register_cleanup($code_ref)
+
+Register a cleanup function which is called just before $r->pool is
+destroyed.
+
+   $r->register_cleanup(sub {
+       my $r = shift;
+       warn "registered cleanup called for ", $r->uri, "\n";
+   });
+
 =back
 
 =head1 CGI SUPPORT
@@ -755,28 +817,28 @@ type of interface.
 Return a %hash that can be used to set up a standard CGI environment.
 Typical usage would be:
 
-  %ENV = $r->cgi_env
+   %ENV = $r->cgi_env
 
 B<NOTE:> The $ENV{GATEWAY_INTERFACE} is set to C<'CGI-Perl/1.1'> so
 you can say:
 
-  if($ENV{GATEWAY_INTERFACE} =~ /^CGI-Perl/) {
-      #do mod_perl stuff
-  }
-  else {
-     #do normal CGI stuff
-  }
+   if($ENV{GATEWAY_INTERFACE} =~ /^CGI-Perl/) {
+       # do mod_perl stuff
+   }
+   else {
+       # do normal CGI stuff
+   }
 
 When given a key => value pair, this will set an environment variable.
 
- $r->cgi_env(REMOTE_GROUP => "camels");
+   $r->cgi_env(REMOTE_GROUP => "camels");
 
 =item $r->cgi_var($key);
 
 Calls $r->cgi_env($key) in a scalar context to prevent the mistake
 of calling in a list context.
 
-    my $doc_root = $r->cgi_env('DOCUMENT_ROOT');
+   my $doc_root = $r->cgi_env('DOCUMENT_ROOT');
 
 =item $r->send_cgi_header()
 
@@ -784,11 +846,11 @@ Take action on certain headers including I<Status:>, I<Location:> and
 I<Content-type:> just as mod_cgi does, then calls
 $r->send_http_header().  Example of use:
 
-  $r->send_cgi_header("
-  Location: /foo/bar
-  Content-type: text/html 
-  
-  ");
+   $r->send_cgi_header(<<EOT);
+   Location: /foo/bar
+   Content-type: text/html 
+   
+   EOT
 
 =back
 
@@ -808,7 +870,7 @@ The request failed, why??  Write a message to the server errorlog.
 
 Uh, oh.  Write a message to the server errorlog.
 
-  $r->log_error("Some text that goes in the error_log");
+   $r->log_error("Some text that goes in the error_log");
 
 =item $r->warn($message)
 
@@ -832,11 +894,13 @@ In opposite to unescape_url it translates the plus sign to space.
 
 =item Apache::perl_hook($hook)
 
-Test to see if a callback hook is enabled
+Returns true if the specified callback hook is enabled:
 
- for (qw(Access Authen Authz Fixup HeaderParser Log Trans Type)) {
-    print "$_ hook enabled\n" if Apache::perl_hook($_);
- }  
+   for (qw(Access Authen Authz ChildInit Cleanup Fixup
+           HeaderParser Init Log Trans Type))
+   {
+       print "$_ hook enabled\n" if Apache::perl_hook($_);
+   }  
 
 =back
 
@@ -845,9 +909,11 @@ Test to see if a callback hook is enabled
 perl(1),
 Apache::Constants(3),
 Apache::Registry(3),
-Apache::CGI(3),
 Apache::Debug(3),
-Apache::Options(3)
+Apache::Options(3),
+CGI::Apache(3)
+
+Apache C API notes at C<http://www.apache.org/docs/>
 
 =head1 AUTHORS
 
