@@ -1,18 +1,15 @@
 package Apache::Constants;
 
-use vars qw($VERSION @ISA @EXPORT);
-
-$VERSION = "1.05";
+$Apache::Constants::VERSION = "1.05";
 
 use Carp ();
 use Exporter ();
 use DynaLoader ();
 use strict;
-use vars qw(@ISA %EXPORT_TAGS @EXPORT_OK @EXPORT);
 
-@ISA = qw(Exporter DynaLoader);
+@Apache::Constants::ISA = qw(Exporter DynaLoader);
 
-%EXPORT_TAGS = (
+%Apache::Constants::EXPORT_TAGS = (
    options => [qw(OPT_NONE OPT_INDEXES OPT_INCLUDES 
 		  OPT_SYM_LINKS OPT_EXECCGI OPT_UNSET OPT_INCNOEXEC
 		  OPT_SYM_OWNER OPT_MULTI OPT_ALL)],
@@ -35,9 +32,10 @@ use vars qw(@ISA %EXPORT_TAGS @EXPORT_OK @EXPORT);
 			 HTTP_SERVICE_UNAVAILABLE
 			 HTTP_VARIANT_ALSO_VARIES)],
     common => [qw(OK DECLINED DONE NOT_FOUND FORBIDDEN AUTH_REQUIRED SERVER_ERROR)],
+    methods => [qw(M_CONNECT M_DELETE M_GET M_INVALID M_OPTIONS M_POST M_PUT M_TRACE)],
 );
 
-@EXPORT_OK = qw(
+@Apache::Constants::EXPORT_OK = qw(
    MOVED
    HTTP_METHOD_NOT_ALLOWED 
    HTTP_NOT_ACCEPTABLE 
@@ -45,9 +43,10 @@ use vars qw(@ISA %EXPORT_TAGS @EXPORT_OK @EXPORT);
    HTTP_PRECONDITION_FAILED
    HTTP_SERVICE_UNAVAILABLE
    HTTP_VARIANT_ALSO_VARIES
+   MODULE_MAGIC_NUMBER
 );
    
-@EXPORT = qw(
+@Apache::Constants::EXPORT = qw(
    AUTH_REQUIRED
    BAD_GATEWAY
    BAD_REQUEST
@@ -59,8 +58,10 @@ use vars qw(@ISA %EXPORT_TAGS @EXPORT_OK @EXPORT);
    M_DELETE
    M_GET
    M_INVALID
+   M_OPTIONS
    M_POST
    M_PUT
+   M_TRACE
    NOT_FOUND
    NOT_IMPLEMENTED
    OK
@@ -82,29 +83,17 @@ use vars qw(@ISA %EXPORT_TAGS @EXPORT_OK @EXPORT);
    USE_LOCAL_COPY
 );
 
-#hmm
-sub MOVED () {301}
-sub AUTH_REQUIRED () {401}
+eval { bootstrap Apache::Constants $Apache::Constants::VERSION; };
+if($@) {
+    my $gw = $ENV{GATEWAY_INTERFACE} || '';
+    die "$@\n" if substr($gw,0,8) eq 'CGI-Perl';
+    warn "warning: can't `bootstrap Apache::Constants $Apache::Constants::VERSION' outside of httpd\n";
+}
 
-{
-    my(%SEEN,@export,$key,$val);
-    while(($key,$val) = each %EXPORT_TAGS) {
-	push @export, grep {!$SEEN{$_}++} @$val;
-    }
-    push @export, grep {!$SEEN{$_}++} @EXPORT;
-    eval { bootstrap Apache::Constants $VERSION; };
-    if($@) {
-	my $gw = $ENV{GATEWAY_INTERFACE} || '';
-	die "$@\n" if substr($gw,0,8) eq 'CGI-Perl';
-	warn "warning: can't `bootstrap Apache::Constants $VERSION' outside of httpd\n";
-    }
-    else {
-	foreach $key (@export) {
-	    next if defined &$key;
-	    $val = constant($key);
-	    eval "sub $key () { $val }";
-	}
-    }
+sub AUTOLOAD {
+                    #why must we stringify first???
+    __AUTOLOAD() if "$Apache::Constants::AUTOLOAD"; 
+    goto &$Apache::Constants::AUTOLOAD;
 }
 
 1;
