@@ -1,4 +1,5 @@
 package Apache::Include;
+use Apache::Constants ':common';
 use Apache::Registry ();
 
 $VERSION = "1.00";
@@ -10,7 +11,8 @@ sub handler {
     $r->uri($uri);
     $r->translate_name; 
     Apache::Registry::handler($r);
-    $r->uri($uri); $r->filename($fname); #reset
+    $r->uri($ouri); $r->filename($fname); #reset
+    OK;
 }
 
 1;
