@@ -3,26 +3,26 @@ use Apache ();
 #use strict; #eval'd scripts will inherit hints
 use Apache::Constants qw(:common &OPT_EXECCGI);
 use FileHandle ();
-use File::Basename qw(dirname);
-use Cwd qw(fastcwd);
+use File::Basename ();
+use Cwd ();
 
-use vars qw($VERSION $Debug);
-#$Id: Registry.pm,v 1.35 1997/07/29 02:11:12 dougm Exp $
-$VERSION = (qw$Revision: 1.35 $)[1];
+#$Id: Registry.pm,v 1.36 1997/09/16 00:47:48 dougm Exp $
+$Apache::Registry::VERSION = (qw$Revision: 1.36 $)[1];
 
-$Debug ||= 0;
+$Apache::Registry::Debug ||= 0;
 # 1 => log recompile in errorlog
 # 2 => Apache::Debug::dump in case of $@
 # 4 => trace pedantically
-Apache->module('Apache::Debug') if $Debug;
+Apache->module('Apache::Debug') if $Apache::Registry::Debug;
 
-my(@cleanup);
 my $Is_Win32 = $^O eq "MSWin32";
 
 sub handler {
     my($r) = @_;
     Apache->request($r);
     my $filename = $r->filename;
+    #local $0 = $filename; #this core dumps!?
+    *0 = \$filename;
     my $oldwarn = $^W;
     $r->log_error("Apache::Registry::handler for $filename in process $$")
 	if $Debug & 4;
@@ -52,6 +52,10 @@ sub handler {
 	    substr($r->uri, 0, length($r->uri)-length($r->path_info)) :
 		$r->uri;
 
+	my $srv = $r->server;
+	$script_name = join "/", $srv->server_hostname, $script_name
+	    if $srv->is_virtual;
+
 	# Escape everything into valid perl identifiers
 	$script_name =~ s/([^A-Za-z0-9\/])/sprintf("_%2x",unpack("C",$1))/eg;
 
@@ -64,12 +68,13 @@ sub handler {
 			  ]egx;
 
 	my $package = "Apache::ROOT$script_name";
+	$Apache::Registry::curstash = $script_name;
 	$r->log_error("Apache::Registry::handler package $package")
 	   if $Debug & 4;
 
 
-	my $cwd = fastcwd;
-	chdir dirname $r->filename;
+	my $cwd = Cwd::fastcwd();
+	chdir File::Basename::dirname($r->filename);
 
 	if (
 	    defined($Apache::Registry->{$package}{mtime})
@@ -82,7 +87,7 @@ sub handler {
 		if $Debug & 4;
 	    my($sub);
 	    {
-		my $fh = new FileHandle $filename;
+		my $fh = FileHandle->new($filename);
 		(my $oldrs, local $/) = ($/, undef);
 		$sub = <$fh>;
 		$sub = parse_cmdline($sub);
@@ -92,6 +97,7 @@ sub handler {
 	    # compile this subroutine into the uniq package name
             $r->log_error("Apache::Registry::handler eval-ing") if $Debug & 4;
  	    undef &{"$package\::handler"} unless $Debug & 4; #avoid warnings
+	    $r->clear_rgy_endav($script_name);
 
 	    my $eval = join(
 			    '',
@@ -124,7 +130,7 @@ sub handler {
 	my $err = $@;
 	if($INC{'CGI.pm'}) {
 	    use Exporter ();
-	    eval { Exporter::require_version('CGI', 2.35); }; 
+	    eval { Exporter::require_version('CGI', 2.36); }; 
 	    $err .= $@ if $@;
 	}
 	if ($err) {
@@ -133,10 +139,6 @@ sub handler {
 	    return Apache::Debug::dump($r, SERVER_ERROR);
 	}
 
-	#run cleanup handlers if any
-	for (@cleanup) { &{$_}($r) }
-	@cleanup = ();
-
 	return $r->status;
     } else {
 	return NOT_FOUND unless $Debug & 2;
@@ -164,6 +166,7 @@ sub parse_cmdline {
     my $sub = shift;
     my($line) = $sub =~ /^(.*)$/m;
     my(@cmdline) = split /\s+/, $line;
+    return $sub unless @cmdline;
     return $sub unless shift(@cmdline) =~ /^\#!/;
     my($s, @s, $prepend);
     $prepend = "";
@@ -180,11 +183,6 @@ sub parse_cmdline {
     return $sub;
 }
 
-sub push_cleanup {
-    my $self = shift;
-    push @cleanup, @_;
-}
-
 #trick so we show up under CPAN/modules/by-module/CGI/
 package CGI::mod_perl;
 
