@@ -50,7 +50,7 @@
  *
  */
 
-/* $Id: mod_perl.c,v 1.59 1997/06/12 00:17:41 dougm Exp $ */
+/* $Id: mod_perl.c,v 1.63 1997/09/16 00:47:48 dougm Exp $ */
 
 /* 
  * And so it was decided the camel should be given magical multi-colored
@@ -61,12 +61,15 @@
 #define CORE_PRIVATE 
 #include "mod_perl.h"
 
+#ifdef MULTITHREAD
+void *mod_perl_mutex = &mod_perl_mutex;
+#else
+void *mod_perl_dummy_mutex = &mod_perl_dummy_mutex;
+#endif
+
 static IV mp_request_rec;
 static int seqno = 0;
 static int perl_is_running = 0;
-static int sent_header = 1;
-static int registered_cleanups = 0;
-static int set_pid = 0;
 static PerlInterpreter *perl = NULL;
 static AV *orig_inc = Nullav;
 static AV *cleanup_av = Nullav;
@@ -98,9 +101,6 @@ static command_rec perl_cmds[] = {
       NULL,
       OR_ALL, TAKE2, "Perl %ENV key and value" },
     { "PerlSendHeader", perl_cmd_sendheader,
-      NULL,
-      OR_ALL, FLAG, "Tell mod_perl to send basic_http_header" },
-    { "PerlNewSendHeader", perl_cmd_new_sendheader,
       NULL,
       OR_ALL, FLAG, "Tell mod_perl to parse and send HTTP headers" },
     { "PerlSetupEnv", perl_cmd_env,
@@ -138,6 +138,15 @@ static command_rec perl_cmds[] = {
 #endif
 #ifdef PERL_HEADER_PARSER
     { PERL_HEADER_PARSER_CMD_ENTRY },
+#endif
+#ifdef PERL_CHILD_INIT
+    { PERL_CHILD_INIT_CMD_ENTRY },
+#endif
+#ifdef PERL_CHILD_EXIT
+    { PERL_CHILD_EXIT_CMD_ENTRY },
+#endif
+#ifdef PERL_POST_READ_REQUEST
+    { PERL_POST_READ_REQUEST_CMD_ENTRY },
 #endif
     { NULL }
 };
@@ -148,12 +157,12 @@ static handler_rec perl_handlers [] = {
     { NULL }
 };
 
-module perl_module = {
+module MODULE_VAR_EXPORT perl_module = {
     STANDARD_MODULE_STUFF,
     perl_startup,                 /* initializer */
-    create_perl_dir_config,    /* create per-directory config structure */
+    perl_create_dir_config,    /* create per-directory config structure */
     perl_merge_dir_config,     /* merge per-directory config structures */
-    create_perl_server_config, /* create per-server config structure */
+    perl_create_server_config, /* create per-server config structure */
     NULL,                      /* merge per-server config structures */
     perl_cmds,                 /* command table */
     perl_handlers,             /* handlers */
@@ -167,6 +176,15 @@ module perl_module = {
 #if MODULE_MAGIC_NUMBER >= 19970103
     PERL_HEADER_PARSER_HOOK,   /* header parser */
 #endif
+#if MODULE_MAGIC_NUMBER >= 19970719
+    PERL_CHILD_INIT_HOOK,   /* child_init */
+#endif
+#if MODULE_MAGIC_NUMBER >= 19970728
+    PERL_CHILD_EXIT_HOOK,   /* child_exit */
+#endif
+#if MODULE_MAGIC_NUMBER >= 19970825
+    PERL_POST_READ_REQUEST_HOOK,   /* post_read_request */
+#endif
 };
 
 #if defined(STRONGHOLD) && !defined(APACHE_SSL)
@@ -178,49 +196,64 @@ int PERL_RUNNING (void)
     return (perl_is_running);
 }
 
+void perl_shutdown (server_rec *s, pool *p)
+{
+    /* execute END blocks we suspended during perl_startup() */
+    perl_run_endav("perl_shutdown"); 
+
+#ifdef HAVE_PERL_5__4
+    perl_destruct_level = 2;
+
+    MP_TRACE(fprintf(stderr, 
+		     "destructing and freeing perl interpreter..."));
+
+    perl_util_cleanup();
+
+    mp_request_rec = 0;
+
+    av_undef(orig_inc);
+    SvREFCNT_dec((SV*)orig_inc);
+    orig_inc = Nullav;
+
+    av_undef(cleanup_av);
+    SvREFCNT_dec((SV*)cleanup_av);
+    cleanup_av = Nullav;
+
+#ifdef PERL_STACKED_HANDLERS
+    hv_undef(stacked_handlers);
+    SvREFCNT_dec((SV*)stacked_handlers);
+    stacked_handlers = Nullhv;
+#endif
+    
+    perl_destruct(perl);
+    perl_free(perl);
+    perl_is_running = 0;
+    MP_TRACE(fprintf(stderr, "ok\n"));
+
+#else
+
+#endif
+}
+
 void perl_startup (server_rec *s, pool *p)
 {
     char *argv[] = { NULL, NULL, NULL, NULL, NULL };
     char *constants[] = { "Apache::Constants", "OK", "DECLINED", NULL };
     int status, i, argc=2, t=0, w=0;
-    perl_server_config *cls;
+    dPSRV(s);
 
+#ifndef WIN32
     argv[0] = server_argv0;
+#endif
 
     if(perl_is_running++) {
-	if(perl_is_running > 2) {
-
-#if 0 /* XXX restarts are _tricky_, it'll work right someday, maybe */
-
-#ifdef HAVE_PERL_5__4
-	    perl_destruct_level = 1;
-#endif
-	    fprintf(stderr, "mod_perl_restart: \n");
-	    if (perl_destruct_level > 0) {
-		MP_TRACE(fprintf(stderr, 
-			 "destructing and freeing perl interpreter...ok\n"));
-		perl_destruct(perl);
-		perl_free(perl);
-	    }
-	    else
-		return;
+#if 0
+      perl_shutdown(s, p);
 #else
-	    MP_TRACE(fprintf(stderr, "perl_startup: perl aleady running...ok\n"));
-	    return;
-#endif /*0*/
-
-	}
-	else {
-	    MP_TRACE(fprintf(stderr, "perl_startup: perl aleady running...ok\n"));
-	    return;
-	}
-    }
-    cls = get_module_config (s->module_config, &perl_module);   
-
-#ifndef PERL_TRACE
-    if (s->error_log)
-	error_log2stderr(s);
+      MP_TRACE(fprintf(stderr, "perl_startup: perl aleady running...ok\n"));
+      return;
 #endif
+    }
 
     MP_TRACE(fprintf(stderr, "allocating perl interpreter..."));
     if((perl = perl_alloc()) == NULL) {
@@ -243,7 +276,8 @@ void perl_startup (server_rec *s, pool *p)
 	argc++;
     }
 
-    argv[1+t+w] = cls->PerlScript;
+    argv[1+t+w] = cls->PerlScript ? 
+	server_root_relative(p, cls->PerlScript) : NULL;
 
     if (argv[1+t+w] == NULL) {
 	argv[1+t+w] = "-e";
@@ -263,7 +297,7 @@ void perl_startup (server_rec *s, pool *p)
     }
     MP_TRACE(fprintf(stderr, "ok\n"));
 
-    perl_clear_env;
+    perl_clear_env();
     hv_store(PerlEnvHV, "GATEWAY_INTERFACE", 17, 
 	     newSVpv(PERL_GATEWAY_INTERFACE,0), 0);
 
@@ -272,14 +306,31 @@ void perl_startup (server_rec *s, pool *p)
 	 * we'll only get here if there's a `use CGI' in the PerlScript file 
 	 */
 	bool old_warn = dowarn;
-	hv_delete(GvHV(incgv), "CGI.pm", 6, G_DISCARD);
+	(void)hv_delete(GvHV(incgv), "CGI.pm", 6, G_DISCARD);
 	MP_TRACE(fprintf(stderr, 
 		 "Reloading CGI.pm so it sees GATEWAY_INTERFACE...\n"));
 	dowarn = FALSE; perl_require_module("CGI", s); dowarn = old_warn;
     }
 
     MP_TRACE(fprintf(stderr, "running perl interpreter..."));
+
+    ENTER;
+    /* suspend END blocks */
+    save_aptr(&endav);
+    endav = Nullav;
+
     status = perl_run(perl);
+
+    MP_TRACE(fprintf(stderr, 
+	     "mod_perl: %d END blocks encountered during server startup\n",
+	     AvFILL(endav)+1));
+#if MODULE_MAGIC_NUMBER < 19970728
+    if(endav)
+	fprintf(stderr, "mod_perl: cannot run END blocks encoutered at server startup without apache_1.3b1+\n");
+#endif
+
+    LEAVE;
+
     if (status != OK) {
 	MP_TRACE(fprintf(stderr,"not ok, status=%d\n", status));
 	perror("run");
@@ -317,21 +368,31 @@ void perl_startup (server_rec *s, pool *p)
     if(!stacked_handlers)
 	stacked_handlers = newHV();
 #endif 
+#ifdef MULTITHREAD
+    mod_perl_mutex = create_mutex(NULL);
+#endif
 }
 
-int mod_perl_sent_header(SV *self, int val)
+int mod_perl_sent_header(request_rec *r, int val)
 {
-    if(val) sent_header = val;
-    return sent_header;
+    dPPDIR;
+
+    if(val) MP_SENTHDR_on(cld);
+    val = MP_SENTHDR(cld) ? 1 : 0;
+    return MP_SENDHDR(cld) ? val : 1;
 }
 
+#ifndef perl_init_ids
+#define perl_init_ids mod_perl_init_ids()
+#endif
+
 int perl_handler(request_rec *r)
 {
-    int status = OK;
-    perl_dir_config *cld = get_module_config (r->per_dir_config,
-					      &perl_module);   
+    dSTATUS;
+    dPPDIR;
 
     (void)perl_request_rec(r); 
+    register_cleanup(r->pool, NULL, mod_perl_end_cleanup, NULL);
 
 #ifdef USE_SFIO
     IoFLAGS(GvIOp(defoutgv)) |= IOf_FLUSH; /* $|=1 */
@@ -339,62 +400,107 @@ int perl_handler(request_rec *r)
     IoFLAGS(GvIOp(defoutgv)) &= ~IOf_FLUSH; /* $|=0 */
 #endif
 
-    MP_TRACE(fprintf(stderr, "(ent): SVs = %5d, OBJs = %5d\n",
-		     sv_count, sv_objcount));
+    MP_TRACE(fprintf(stderr, "perl_handler ENTER: SVs = %5d, OBJs = %5d\n",
+		     (int)sv_count, (int)sv_objcount));
     ENTER;
     SAVETMPS;
 
+    save_aptr(&endav); 
+    endav = Nullav;
+
     /* hookup STDIN & STDOUT to the client */
     perl_stdout2client(r);
     perl_stdin2client(r);
 
-#ifndef PERL_TRACE
-    /* hookup STDERR to the error_log */
-    if (r->server->error_log)
-	error_log2stderr(r->server);
-#endif
-
     seqno++;
-    register_cleanup(r->pool, NULL, mod_perl_end_cleanup, NULL);
 
-    /*don't do anything special unless PerlNewSendHeader*/ 
-    sent_header = (cld->new_sendheader ? 0 : 1); 
-    if(cld->sendheader) {
-	MP_TRACE(fprintf(stderr, "mod_perl sending basic_http_header...\n"));
-	basic_http_header(r);
-    }
-    if(cld->setup_env) 
+    if(MP_ENV(cld)) 
 	perl_setup_env(r);
 
-    mod_perl_dir_env(cld);
-
     PERL_CALLBACK("PerlHandler", cld->PerlHandler);
 
+    perl_run_rgy_endav(r->uri);
+
     FREETMPS;
     LEAVE;
-    MP_TRACE(fprintf(stderr, "(lea): SVs = %5d, OBJs = %5d\n", 
-		     sv_count, sv_objcount));
+    MP_TRACE(fprintf(stderr, "perl_handler LEAVE: SVs = %5d, OBJs = %5d\n", 
+		     (int)sv_count, (int)sv_objcount));
     return status;
 }
 
+
+#ifdef PERL_CHILD_INIT
+void PERL_CHILD_INIT_HOOK(server_rec *s, pool *p)
+{
+    request_rec *r = (request_rec *)palloc(p, sizeof(request_rec));
+    dSTATUS;
+    dPSRV(s);
+
+    r->pool = p; 
+    r->server = s;
+
+    mod_perl_init_ids();
+
+    PERL_CALLBACK("PerlChildInitHandler", cls->PerlChildInitHandler);
+}
+#endif
+
+#ifdef PERL_CHILD_EXIT
+void PERL_CHILD_EXIT_HOOK(server_rec *s, pool *p)
+{
+    request_rec *r = (request_rec *)palloc(p, sizeof(request_rec));
+    dSTATUS;
+    dPSRV(s);
+
+    r->pool = p; 
+    r->server = s;
+
+    PERL_CALLBACK("PerlChildExitHandler", cls->PerlChildExitHandler);
+
+    perl_shutdown(s,p);
+}
+#endif
+
+#ifdef PERL_POST_READ_REQUEST
+int PERL_POST_READ_REQUEST_HOOK(request_rec *r)
+{
+    dSTATUS;
+    dPSRV(r->server);
+    PERL_CALLBACK("PerlPostReadRequestHandler", cls->PerlPostReadRequestHandler);
+    return status;
+}
+#endif
+
 #ifdef PERL_TRANS
 int PERL_TRANS_HOOK(request_rec *r)
 {
-    int status = DECLINED;
-    perl_server_config *cls = get_module_config (r->server->module_config,
-						 &perl_module);   
+    dSTATUS;
+    dPSRV(r->server);
     PERL_CALLBACK("PerlTransHandler", cls->PerlTransHandler);
     return status;
 }
 #endif
 
+#ifdef PERL_HEADER_PARSER
+int PERL_HEADER_PARSER_HOOK(request_rec *r)
+{
+    dSTATUS;
+    dPPDIR;
+#ifdef PERL_INIT
+    PERL_CALLBACK("PerlInitHandler", 
+			 cld->PerlInitHandler);
+#endif
+    PERL_CALLBACK("PerlHeaderParserHandler", 
+			 cld->PerlHeaderParserHandler);
+    return status;
+}
+#endif
+
 #ifdef PERL_AUTHEN
 int PERL_AUTHEN_HOOK(request_rec *r)
 {
-    int status = DECLINED;
-    perl_dir_config *cld = get_module_config (r->per_dir_config,
-					      &perl_module);   
-    mod_perl_dir_env(cld);
+    dSTATUS;
+    dPPDIR;
     PERL_CALLBACK("PerlAuthenHandler", cld->PerlAuthenHandler);
     return status;
 }
@@ -403,10 +509,8 @@ int PERL_AUTHEN_HOOK(request_rec *r)
 #ifdef PERL_AUTHZ
 int PERL_AUTHZ_HOOK(request_rec *r)
 {
-    int status = DECLINED;
-    perl_dir_config *cld = get_module_config (r->per_dir_config,
-					      &perl_module);   
-    mod_perl_dir_env(cld);
+    dSTATUS;
+    dPPDIR;
     PERL_CALLBACK("PerlAuthzHandler", cld->PerlAuthzHandler);
     return status;
 }
@@ -415,10 +519,8 @@ int PERL_AUTHZ_HOOK(request_rec *r)
 #ifdef PERL_ACCESS
 int PERL_ACCESS_HOOK(request_rec *r)
 {
-    int status = DECLINED;
-    perl_dir_config *cld = get_module_config (r->per_dir_config,
-					      &perl_module);   
-    mod_perl_dir_env(cld);
+    dSTATUS;
+    dPPDIR;
     PERL_CALLBACK("PerlAccessHandler", cld->PerlAccessHandler);
     return status;
 }
@@ -427,10 +529,8 @@ int PERL_ACCESS_HOOK(request_rec *r)
 #ifdef PERL_TYPE
 int PERL_TYPE_HOOK(request_rec *r)
 {
-    int status = DECLINED;
-    perl_dir_config *cld = get_module_config (r->per_dir_config,
-					      &perl_module);   
-    mod_perl_dir_env(cld);
+    dSTATUS;
+    dPPDIR;
     PERL_CALLBACK("PerlTypeHandler", cld->PerlTypeHandler);
     return status;
 }
@@ -439,10 +539,8 @@ int PERL_TYPE_HOOK(request_rec *r)
 #ifdef PERL_FIXUP
 int PERL_FIXUP_HOOK(request_rec *r)
 {
-    int status = DECLINED;
-    perl_dir_config *cld = get_module_config (r->per_dir_config,
-					      &perl_module);   
-    mod_perl_dir_env(cld);
+    dSTATUS;
+    dPPDIR;
     PERL_CALLBACK("PerlFixupHandler", cld->PerlFixupHandler);
     return status;
 }
@@ -451,10 +549,9 @@ int PERL_FIXUP_HOOK(request_rec *r)
 #ifdef PERL_LOG
 int PERL_LOG_HOOK(request_rec *r)
 {
-    int status = DECLINED, rstatus;
-    perl_dir_config *cld = get_module_config (r->per_dir_config,
-					      &perl_module);   
-    mod_perl_dir_env(cld);
+    dSTATUS;
+    dPPDIR;
+    int rstatus;
     PERL_CALLBACK("PerlLogHandler", cld->PerlLogHandler);
     rstatus = status;
 #ifdef PERL_CLEANUP
@@ -466,7 +563,8 @@ int PERL_LOG_HOOK(request_rec *r)
 
 void mod_perl_end_cleanup(void *data)
 {
-    perl_clear_env;
+    (void)acquire_mutex(mod_perl_mutex); 
+    perl_clear_env();
     av_undef(GvAV(incgv));
     SvREFCNT_dec(GvAV(incgv));
     GvAV(incgv) = Nullav;
@@ -474,6 +572,7 @@ void mod_perl_end_cleanup(void *data)
     /* reset $/ */
     sv_setpvn(GvSV(gv_fetchpv("/", FALSE, SVt_PV)), "\n", 1);
     MP_TRACE(fprintf(stderr, "perl_end_cleanup...ok\n"));
+    (void)release_mutex(mod_perl_mutex); 
 }
 
 void mod_perl_cleanup_handler(void *data)
@@ -482,32 +581,16 @@ void mod_perl_cleanup_handler(void *data)
     SV *cv;
     I32 i;
 
+    (void)acquire_mutex(mod_perl_mutex); 
     MP_TRACE(fprintf(stderr, "running registered cleanup handlers...\n")); 
     for(i=0; i<=AvFILL(cleanup_av); i++) { 
 	cv = *av_fetch(cleanup_av, i, 0);
 	perl_call_handler(cv, (request_rec *)r, Nullav);
     }
-    registered_cleanups = 0;
     av_clear(cleanup_av);
+    (void)release_mutex(mod_perl_mutex); 
 }
 
-#ifdef PERL_HEADER_PARSER
-int PERL_HEADER_PARSER_HOOK(request_rec *r)
-{
-    int status = DECLINED;
-    perl_dir_config *cld = get_module_config (r->per_dir_config,
-					      &perl_module);   
-    mod_perl_dir_env(cld);
-#ifdef PERL_INIT
-    PERL_CALLBACK("PerlInitHandler", 
-			 cld->PerlInitHandler);
-#endif
-    PERL_CALLBACK("PerlHeaderParserHandler", 
-			 cld->PerlHeaderParserHandler);
-    return status;
-}
-#endif
-
 #ifdef PERL_METHOD_HANDLERS
 int perl_handler_ismethod(HV *class, char *sub)
 {
@@ -533,54 +616,56 @@ int perl_handler_ismethod(HV *class, char *sub)
 
 void mod_perl_register_cleanup(request_rec *r, SV *sv)
 {
-    if(!registered_cleanups) {
+    dPPDIR;
+
+    if(!MP_RCLEANUP(cld)) {
 	(void)perl_request_rec(r); 
 	register_cleanup(r->pool, (void*)r,
 			 mod_perl_cleanup_handler, NULL);
-	++registered_cleanups;
+	MP_RCLEANUP_on(cld);
 	if(cleanup_av == Nullav) cleanup_av = newAV();
     }
     MP_TRACE(fprintf(stderr, "registering PerlCleanupHandler\n"));
     
-    av_push(cleanup_av, SvREFCNT_inc(sv));
+    ++SvREFCNT(sv); av_push(cleanup_av, sv);
 }
 
 #ifdef PERL_STACKED_HANDLERS
 
-int mod_perl_push_handlers(SV *self, SV *hook, SV *sub, AV *handlers)
+int mod_perl_push_handlers(SV *self, char *hook, SV *sub, AV *handlers)
 {
-    char *key = SvPV(hook,na);
-    int do_store=0;
+    int do_store=0, len=strlen(hook);
     SV **svp;
 
     if(self && SvTRUE(sub)) {
 	if(handlers == Nullav) {
-	    svp = hv_fetch(stacked_handlers, key, SvCUR(hook), 0);
-	    MP_TRACE(fprintf(stderr, "fetching %s stack\n", key));
+	    svp = hv_fetch(stacked_handlers, hook, len, 0);
+	    MP_TRACE(fprintf(stderr, "fetching %s stack\n", hook));
 	    if(svp && SvTRUE(*svp) && SvROK(*svp)) {
 		handlers = (AV*)SvRV(*svp);
 	    }
 	    else {
-		MP_TRACE(fprintf(stderr, "%s handlers stack undef, creating\n", key));
+		MP_TRACE(fprintf(stderr, "%s handlers stack undef, creating\n", hook));
 		handlers = newAV();
 	    }
 	    do_store = 1;
 	}
 	    
 	if(SvROK(sub) && (SvTYPE(SvRV(sub)) == SVt_PVCV)) {
-	    MP_TRACE(fprintf(stderr, "pushing CODE ref into `%s' handlers\n", key));
+	    MP_TRACE(fprintf(stderr, "pushing CODE ref into `%s' handlers\n", hook));
 	}
 	else if(SvPOK(sub)) {
 	    MP_TRACE(fprintf(stderr, "pushing `%s' into `%s' handlers\n", 
-		   SvPV(sub,na), key));
+		   SvPV(sub,na), hook));
 	}
 	else {
 	    warn("mod_perl_push_handlers: Not a subroutine name or CODE reference!");
 	}
 
-	av_push(handlers, SvREFCNT_inc(sub));
+	++SvREFCNT(sub); av_push(handlers, sub);
+
 	if(do_store) 
-	    hv_store(stacked_handlers, key, SvCUR(hook), 
+	    hv_store(stacked_handlers, hook, len, 
 		     (SV*)newRV((SV*)handlers), 0);
 	return 1;
     }
@@ -589,21 +674,40 @@ int mod_perl_push_handlers(SV *self, SV *hook, SV *sub, AV *handlers)
 
 int perl_run_stacked_handlers(char *hook, request_rec *r, AV *handlers)
 {
-    int status=DECLINED, do_clear=0;
-    I32 i;
+    dSTATUS;
+    I32 i, do_clear=FALSE;
     SV *sub, **svp; 
     int hook_len = strlen(hook);
 
     if(handlers == Nullav) {
-	svp = hv_fetch(stacked_handlers, hook, hook_len, 0);
-	if(!svp || !SvTRUE(*svp) || !SvROK(*svp)) {
+	if(hv_exists(stacked_handlers, hook, hook_len)) {
+	   svp = hv_fetch(stacked_handlers, hook, hook_len, 0);
+	   if(svp && SvROK(*svp)) 
+	       handlers = (AV*)SvRV(*svp);
+	}
+	else {
 	    MP_TRACE(fprintf(stderr, "`%s' push_handlers() stack is empty\n", hook));
 	    return DECLINED;
 	}
-	handlers = (AV*)SvRV(*svp);
-	do_clear = 1;
+	do_clear = TRUE;
+	MP_TRACE(fprintf(stderr, 
+		 "running %d pushed (stacked) handlers for %s...\n", 
+			 AvFILL(handlers)+1, r->uri)); 
+    }
+    else {
+#ifdef PERL_STACKED_HANDLERS
+      /* XXX: bizarre, 
+	 I only see this with httpd.conf.pl and PerlAccessHandler */
+	if(SvTYPE((SV*)handlers) != SVt_PVAV) {
+	    fprintf(stderr, "%s stack is not an ARRAY!\n", hook);
+	    sv_dump((SV*)handlers);
+	    return DECLINED;
+	}
+#endif
+	MP_TRACE(fprintf(stderr, 
+		 "running %d server configured stacked handlers for %s...\n", 
+			 AvFILL(handlers)+1, r->uri)); 
     }
-
     for(i=0; i<=AvFILL(handlers); i++) {
 	MP_TRACE(fprintf(stderr, "calling &{%s->[%d]}\n", hook, (int)i));
 
@@ -631,8 +735,45 @@ int perl_run_stacked_handlers(char *hook, request_rec *r, AV *handlers)
 
 #endif /* PERL_STACKED_HANDLERS */
 
+/* things to do once per-request */
+void perl_per_request_init(request_rec *r)
+{
+    dPPDIR;
+
+    /* hookup stderr to error_log */
+#ifndef PERL_TRACE
+    if(!MP_DSTDERR(cld)) {
+	if(r->server->error_log) {
+	    error_log2stderr(r->server);
+	    MP_DSTDERR_on(cld);
+	}
+    }
+#endif
+
+    /* set $$, $>, etc., if 1.3a1+, this really happens during child_init */
+    perl_init_ids; 
+
+    /* PerlSetEnv */
+    mod_perl_dir_env(cld);
+
+    /* PerlSendHeader */
+    if(MP_SENDHDR(cld))
+	MP_SENTHDR_off(cld);
+    else
+	MP_SENTHDR_on(cld);
+
+    /* SetEnv PERL5LIB */
+    if(!MP_INCPUSH(cld)) {
+	char *path = table_get(r->subprocess_env, "PERL5LIB");
+	if(path) {
+	    perl_incpush(path);
+	    MP_INCPUSH_on(cld);
+	}
+    }
+}
+
 /* XXX this still needs work, getting there... */
-int perl_call_handler(SV *sv, request_rec *r, AV *args)
+API_EXPORT(int) perl_call_handler(SV *sv, request_rec *r, AV *args)
 {
     int count, status, is_method=0;
     dSP;
@@ -641,6 +782,9 @@ int perl_call_handler(SV *sv, request_rec *r, AV *args)
     CV *cv = Nullcv;
     char *method = "handler";
     int defined_sub = 0, anon = 0;
+      
+    if(r->per_dir_config)
+	perl_per_request_init(r);
 
     if(SvTYPE(sv) == SVt_PV) {
 	char *imp = pstrdup(r->pool, (char *)SvPV(class,na));
@@ -769,9 +913,6 @@ callback:
     }
     PUTBACK;
     
-    /* reset $$ */
-    perl_set_pid;
-
     /* use G_EVAL so we can trap errors */
 #ifdef PERL_METHOD_HANDLERS
     if(is_method)
@@ -795,18 +936,21 @@ callback:
 	if((status == 1) || (status == 200) || (status > 600)) 
 	    status = OK; 
     }
+
     PUTBACK;
     FREETMPS;
     LEAVE;
     MP_TRACE(fprintf(stderr, "perl_call_handler: SVs = %5d, OBJs = %5d\n", 
-	    sv_count, sv_objcount));
+	    (int)sv_count, (int)sv_objcount));
+
+    if(SvMAGICAL(GvSV(errgv)))
+       sv_unmagic(GvSV(errgv), 'U'); /* Apache::exit was called */
+
     return status;
 }
 
 request_rec *perl_request_rec(request_rec *r)
 {
-    /* This will depreciate */
-    /* CTRACE(stderr, "perl_request_rec\n"); */
     if(r != NULL) {
 	mp_request_rec = (IV)r;
 	return NULL;
@@ -819,39 +963,11 @@ SV *perl_bless_request_rec(request_rec *r)
 {
     SV *sv = sv_newmortal();
     sv_setref_pv(sv, "Apache", (void*)r);
-    MP_TRACE(fprintf(stderr, "blessing request_rec\n"));
+    MP_TRACE(fprintf(stderr, "blessing request_rec=(0x%lx)\n",
+		     (unsigned long)r));
     return sv;
 }
 
-int perl_eval_ok(server_rec *s)
-{
-    SV *sv;
-    sv = GvSV(gv_fetchpv("@", TRUE, SVt_PV));
-    if(SvTRUE(sv)) {
-	MP_TRACE(fprintf(stderr, "perl_eval error: %s\n", SvPV(sv,na)));
-	log_error(SvPV(sv, na), s);
-	return -1;
-    }
-    return 0;
-}
-
-int perl_require_module(char *mod, server_rec *s)
-{
-    SV *sv = sv_newmortal();
-    SV *m = newSVpv(mod,0);
-    MP_TRACE(fprintf(stderr, "loading perl module '%s'...", mod)); 
-    sv_setpv(sv, "require ");
-    sv_catsv(sv, m);
-    perl_eval_sv(sv, G_DISCARD);
-    SvREFCNT_dec(m);
-    if(perl_eval_ok(s) != OK) {
-	MP_TRACE(fprintf(stderr, "not ok\n"));
-	return -1;
-    }
-    MP_TRACE(fprintf(stderr, "ok\n"));
-    return 0;
-}
-
 void perl_setup_env(request_rec *r)
 { 
     int klen;
@@ -873,112 +989,9 @@ void perl_setup_env(request_rec *r)
     MP_TRACE(fprintf(stderr, "perl_setup_env...%d keys\n", i));
 }
 
-#ifdef USE_SFIO
-
-typedef struct {
-    Sfdisc_t     disc;   /* the sfio discipline structure */
-    request_rec	*r;
-} Apache_t;
-
-static int sfapachewrite(f, buffer, n, disc)
-    Sfio_t* f;      /* stream involved */
-    char*           buffer;    /* buffer to read into */
-    int             n;      /* number of bytes to send */
-    Sfdisc_t*       disc;   /* discipline */        
+int mod_perl_seqno(SV *self)
 {
-    /* feed buffer to Apache->print */
-    CV *cv = GvCV(gv_fetchpv("Apache::print", FALSE, SVt_PVCV));
-    dSP;
-    ENTER;
-    SAVETMPS;
-    PUSHMARK(sp);
-    XPUSHs(perl_bless_request_rec(((Apache_t*)disc)->r));
-    XPUSHs(sv_2mortal(newSVpv(buffer,n)));
-    PUTBACK;
-    (void)(*CvXSUB(cv))(cv); 
-    FREETMPS;
-    LEAVE;
-    return n;
-}
-
-static int sfapacheread(f, buffer, bufsiz, disc)
-    Sfio_t* f;      /* stream involved */
-    char*           buffer;    /* buffer to read into */
-    int             bufsiz;      /* number of bytes to read */
-    Sfdisc_t*       disc;   /* discipline */        
-{
-    long nrd;
-    request_rec	*r = ((Apache_t*)disc)->r;
-    MP_TRACE(fprintf(stderr, "sfapacheread: want %d bytes\n", bufsiz)); 
-    PERL_READ_FROM_CLIENT;
-    return bufsiz;
-}
-
-Sfdisc_t * sfdcnewapache(request_rec *r)
-{
-    Apache_t*   disc;
-    
-    if(!(disc = (Apache_t*)malloc(sizeof(Apache_t))) )
-	return (Sfdisc_t *)disc;
-    MP_TRACE(fprintf(stderr, "sfdcnewapache(r)\n"));
-    disc->disc.readf   = (Sfread_f)sfapacheread; 
-    disc->disc.writef  = (Sfwrite_f)sfapachewrite;
-    disc->disc.seekf   = (Sfseek_f)NULL;
-    disc->disc.exceptf = (Sfexcept_f)NULL;
-    disc->r = r;
-    return (Sfdisc_t *)disc;
-}
-#endif
-
-void perl_stdout2client(request_rec *r)
-{
-#ifdef USE_SFIO
-    sfdisc(PerlIO_stdout(), SF_POPDISC);
-    sfdisc(PerlIO_stdout(), sfdcnewapache(r));
-#else
-    GV *handle = gv_fetchpv("STDOUT", TRUE, SVt_PVIO);  
-
-#if 0 
-/* XXX so Perl*Handler's can re-tie before PerlHandler is run? 
- * then they'd also be reponsible for re-tie'ing to `Apache'
- * after all PerlHandlers are run, hmm must think.
- */
-    MAGIC *mg;
-    if (SvMAGICAL(handle) && (mg = mg_find((SV*)handle, 'q'))) {
-	char *package = HvNAME(SvSTASH((SV*)SvRV(mg->mg_obj)));
-	if(!strEQ(package, "Apache")) {
-	    fprintf(stderr, "%s tied to %s\n", GvNAME(handle), package);
-	    return;
-	}
-    }
-#endif
-
-    MP_TRACE(fprintf(stderr, "tie *STDOUT => Apache\n"));
-    sv_unmagic((SV*)handle, 'q');
-    sv_magic((SV *)handle, 
-	     (SV *)perl_bless_request_rec(r),
-	     'q', Nullch, 0);
-#endif
-}
-
-void perl_stdin2client(request_rec *r)
-{
-#ifdef USE_SFIO
-    sfdisc(PerlIO_stdin(), SF_POPDISC);
-    sfdisc(PerlIO_stdin(), sfdcnewapache(r));
-    sfsetbuf(PerlIO_stdin(), NULL, 0);
-#else
-    GV *handle = gv_fetchpv("STDIN", TRUE, SVt_PVIO);  
-    MP_TRACE(fprintf(stderr, "tie *STDIN => Apache\n"));
-    sv_unmagic((SV*)handle, 'q');
-    sv_magic((SV *)handle,
-	     (SV *)perl_bless_request_rec(r),
-	     'q', Nullch, 0);
-#endif
-}
-
-int mod_perl_seqno(void)
-{
-    return seqno;
+    if(SvTRUE(self)) return seqno;
+    else return -1;
 }
 
